/**      Cartman2lite.C by Alexander Myasnikov                     **/
/**      Lite version of Cartman block cipher 2.DDP.3              **/
/**      Cartman LITE encrypts 128 bit block with 320 bit key      **/
/**      Gives much better transforming with new key schedule      **/
/**      Freeware for any use,  non-patented, opensource cipher    **/
/**      WEB:       www.darksoftware.narod.ru                      **/


#include <stdlib.h>
#include <string.h>

typedef unsigned long u32;
typedef unsigned char u8;
#define keysizebits     320
#define keysizebytes    keysizebits / 8
#define keysize         keysizebytes / 4
#define subkeysize      keysize / 2
#define ror32(N, R)    _lrotr(N, R)
#define rol32(N, R)    _lrotl(N, R)
#define rounds_total subkeysize * subkeysize

typedef union
{
	u32 d32;
	u8 d8[4];
} r32;

u32 idx_box1[subkeysize];
u32 idx_box2[subkeysize];
u32 idx_box3[subkeysize];
u32 idx_box4[subkeysize];
u32 rot_box2[subkeysize - 1];
u32 rot_box3[subkeysize - 1];
u32 rot_box[keysize];

union
{
	u8 d8[keysizebytes];
	u32 d32[keysize];
}
key;

u32 subkey1[subkeysize];
u32 subkey2[subkeysize];
u32 sbidx1[subkeysize][subkeysize];
u32 sbidx2[subkeysize][subkeysize];
u32 idx1null;
u32 idx2null;
u32 idx3null;

typedef struct
{
	u32 idx1;
	u32 idx2;
	u32 idx3;
	u32 idx4;
	u32 idx5;
	u32 idx6;
	u32 idx5_1;
	u32 idx6_1;
	u32 k1;
	u32 k2;
	u32 k3;
	u32 k4;
	u32 rot1;
	u32 rot2;
	u32 rot_box2;
	u32 rot_box3;

}
idxblock;

idxblock prefetch[rounds_total];

#include "SBox.h"

void
makebox ()
{
	u32 i, n, cnt, x;
	u32 pi, idx1, idx2, idx3, idx4, idx5, idx6;

	idx1null = 0;
	idx2null = 0;
	idx3null = 0;


	for (i = 0; i != subkeysize; i++)
	{
		idx_box1[i] =
			(ror32
			(key.d32[i] + ((key.d32[i] << 4) ^ (key.d32[i] >> 5)),
			(i % 31) + 1)) % (subkeysize - 1);

		idx1null = (idx1null * (idx1null + 1) >> 1) + key.d32[i];
	}

	for (i = subkeysize; i != keysize; i++)
	{
		idx_box2[(keysize - 1) - i] =
			(rol32
			(key.d32[i] + ((key.d32[i] << 4) ^ (key.d32[i] >> 5)),
			(i % 31) + 1)) % (subkeysize - 1);
		idx2null = (idx2null * (idx2null + 1) >> 1) + key.d32[i];
	}

	for (i = 0; i != keysize; i++)
	{
		rot_box[i] =
			((rol32 (key.d32[i] ^ (i + 1), (((keysize - 1) - i) % 31) + 1)) %
			31) + 1;
		idx3null += key.d32[i] * (i + 1);
	}

	idx1null = (idx1null ^ (idx1null >> 16)) % subkeysize;
	idx2null = (idx2null ^ (idx2null >> 16)) % subkeysize;
	idx3null = (idx3null ^ (idx3null >> 16)) % subkeysize;

	for (i = 0; i < subkeysize - 1; i++)
	{
		rot_box2[i] = (subkey2[(subkeysize - 1) - i] % 31) + 1;
		rot_box3[i] = (subkey1[(subkeysize - 1) - i] % 31) + 1;
	}


	for (i = 0; i != subkeysize; i++)
	{
		idx_box3[i] =
			(rol32
			(key.d32[i] + ((key.d32[i] << 5) ^ (key.d32[i] >> 9)),
			(i % 31) + 1)) % (subkeysize - 1);
	}

	for (i = subkeysize; i != keysize; i++)
	{
		idx_box4[(keysize - 1) - i] =
			(ror32
			(key.d32[i] + ((key.d32[i] << 5) ^ (key.d32[i] >> 9)),
			(i % 31) + 1)) % (subkeysize - 1);
	}



	cnt = 0;

	for (i = 0; i != subkeysize; i++)
		for (n = 0; n != subkeysize; n++)
		{
			sbidx1[i][n] =
				(rol32 (key.d8[cnt], i) * (ror32 (key.d8[cnt], (n + 1)) + 1) +
				cnt) % (subkeysize - 1);
			cnt++;
		}

		cnt = 0;
		x = 0;


		for (n = 0; n != subkeysize; n++)
			for (i = 0; i != subkeysize; i++)
			{

				x = ((x << n) ^ (x >> i)) + rol32 (key.d8[cnt], (cnt % 31) + 1) + cnt;
				sbidx2[i][n] = x % (subkeysize - 1);
				cnt++;
			}


			pi = 0;

			for (n = 0; n != subkeysize; n++)
			{
				idx1 = idx_box1[n];
				idx2 = idx_box2[n];
				idx3 = idx_box3[n];
				idx4 = idx_box4[n];


				for (i = 0; i != subkeysize; i++, pi++)
				{

					idx5 = sbidx1[n][i];
					idx6 = sbidx2[n][i];

					prefetch[pi].idx1 = idx1;
					prefetch[pi].idx2 = idx2;
					prefetch[pi].idx3 = idx3;
					prefetch[pi].idx4 = idx4;
					prefetch[pi].idx5 = idx5;
					prefetch[pi].idx6 = idx6;
					prefetch[pi].idx5_1 = idx5 + 1;
					prefetch[pi].idx6_1 = idx6 + 1;


					prefetch[pi].k1 = subkey1[idx1];
					prefetch[pi].k2 = subkey1[idx1 + 1];
					prefetch[pi].k3 = subkey2[idx2];
					prefetch[pi].k4 = subkey2[idx2 + 1];

					prefetch[pi].rot1 = rot_box[idx1 + idx2];
					prefetch[pi].rot2 = rot_box[idx1 + idx2 + 1];

					prefetch[pi].rot_box2 = rot_box2[idx3];
					prefetch[pi].rot_box3 = rot_box3[idx4];



					if (idx1 != (subkeysize - 2))
					{
						idx1++;
					}
					else
					{
						idx1 = 0;
					}

					if (idx2 != (subkeysize - 2))
					{
						idx2++;
					}
					else
					{
						idx2 = 0;
					}


					if (idx3 != (subkeysize - 2))
					{
						idx3++;
					}
					else
					{
						idx3 = 0;
					}

					if (idx4 != (subkeysize - 2))
					{
						idx4++;
					}
					else
					{
						idx4 = 0;
					}


				}
			}



}


void __stdcall
setup (u32 * k)
{
	memmove (&key.d32[0], &k[0], keysizebytes);
	memmove (&subkey1[0], &key.d32[0], subkeysize * 4);
	memmove (&subkey2[0], &key.d32[subkeysize], subkeysize * 4);
	makebox ();

}

void __stdcall
killkey ()
{
	memset (&key.d8[0], 0x00, keysizebytes);
	memset (&prefetch, 0x00, sizeof (prefetch));
}

void __stdcall
crypt (r32 * v)
{
	u32 i, n, v0, v1, j, m, s, t;
	u32 pi = 0;

	for (n = 0; n != subkeysize; n++)
	{



		for (i = 0; i != subkeysize; i++, pi++)
		{



			s = v[0].d32 ^ v[1].d32 ^ v[2].d32 ^ v[3].d32;



			m = s & 3;
			t = v[0].d32;
			v[0].d32 = v[m].d32;
			v[m].d32 = t;

			m = (s >> 8) & 3;
			t = v[1].d32;
			v[1].d32 = v[m].d32;
			v[m].d32 = t;

			m = (s >> 16) & 3;
			t = v[2].d32;
			v[2].d32 = v[m].d32;
			v[m].d32 = t;

			m = (s >> 24) & 3;
			t = v[3].d32;
			v[3].d32 = v[m].d32;
			v[m].d32 = t;


			v[0].d8[0] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[0]];
			v[0].d8[1] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[1]];
			v[0].d8[2] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[2]];
			v[0].d8[3] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[3]];

			v[1].d8[0] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[0]];
			v[1].d8[1] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[1]];
			v[1].d8[2] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[2]];
			v[1].d8[3] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[3]];

			v[2].d8[0] =
				SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[0]];
			v[2].d8[1] =
				SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[1]];
			v[2].d8[2] =
				SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[2]];
			v[2].d8[3] =
				SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[3]];

			v[3].d8[0] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[0]];
			v[3].d8[1] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[1]];
			v[3].d8[2] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[2]];
			v[3].d8[3] =
				SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[3]];


			if (prefetch[pi].idx1 == idx1null)
			{
				v[0].d32 +=
					(v[3].d32 +
					ror32 ((v[1].d32 + ((v[1].d32 << 6) ^ (v[1].d32 >> 9))),
					prefetch[pi].rot1) + prefetch[pi].k1);
			}
			else
			{
				v[0].d32 +=
					(v[3].d32 +
					ror32 ((v[1].d32 + ((v[1].d32 << 6) ^ (v[1].d32 >> 9))) +
					prefetch[pi].k1, prefetch[pi].rot1));
			}

			if (prefetch[pi].idx2 == idx2null)
			{
				v[1].d32 +=
					(v[2].d32 +
					ror32 ((v[0].d32 + ((v[0].d32 << 6) ^ (v[0].d32 >> 9))),
					prefetch[pi].rot2) + prefetch[pi].k2);
			}
			else
			{
				v[1].d32 +=
					(v[2].d32 +
					ror32 ((v[0].d32 + ((v[0].d32 << 6) ^ (v[0].d32 >> 9))) +
					prefetch[pi].k2, prefetch[pi].rot2));
			}


			if (i == idx3null)
			{
				v[2].d32 +=
					rol32 (((v[0].d32 + ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
					prefetch[pi].k3) * 9, prefetch[pi].rot_box2);
				v[3].d32 +=
					rol32 (((v[1].d32 + ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
					prefetch[pi].k4) * 9, prefetch[pi].rot_box3);
			}
			else
			{
				v[2].d32 ^=
					rol32 (((v[0].d32 + ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
					prefetch[pi].k3) * 9, prefetch[pi].rot_box2);
				v[3].d32 ^=
					rol32 (((v[1].d32 + ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
					prefetch[pi].k4) * 9, prefetch[pi].rot_box3);
			}


			v0 = v[0].d32;
			v1 = v[1].d32;
			v[0].d32 = v[2].d32;
			v[1].d32 = v[3].d32;
			v[2].d32 = v0;
			v[3].d32 = v1;






		}
	}
}

void __stdcall
decrypt (r32 * v)
{
	u32 v0, v1, m, s, t;
	u32 pi = rounds_total - 1;

	long n, i, j;

	for (n = subkeysize - 1; n != -1; n--)

	{


		for (i = subkeysize - 1; i != -1; i--, pi--)
		{


			v0 = v[0].d32;
			v1 = v[1].d32;
			v[0].d32 = v[2].d32;
			v[1].d32 = v[3].d32;
			v[2].d32 = v0;
			v[3].d32 = v1;


			if (i == idx3null)
			{
				v[3].d32 -=
					rol32 (((v[1].d32 + ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
					prefetch[pi].k4) * 9, prefetch[pi].rot_box3);
				v[2].d32 -=
					rol32 (((v[0].d32 + ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
					prefetch[pi].k3) * 9, prefetch[pi].rot_box2);
			}
			else
			{
				v[3].d32 ^=
					rol32 (((v[1].d32 + ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
					prefetch[pi].k4) * 9, prefetch[pi].rot_box3);
				v[2].d32 ^=
					rol32 (((v[0].d32 + ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
					prefetch[pi].k3) * 9, prefetch[pi].rot_box2);
			}



			if (prefetch[pi].idx2 == idx2null)
			{
				v[1].d32 -=
					(v[2].d32 +
					ror32 ((v[0].d32 + ((v[0].d32 << 6) ^ (v[0].d32 >> 9))),
					prefetch[pi].rot2) + prefetch[pi].k2);
			}
			else
			{
				v[1].d32 -=
					(v[2].d32 +
					ror32 ((v[0].d32 + ((v[0].d32 << 6) ^ (v[0].d32 >> 9))) +
					prefetch[pi].k2, prefetch[pi].rot2));
			}



			if (prefetch[pi].idx1 == idx1null)
			{
				v[0].d32 -=
					(v[3].d32 +
					ror32 ((v[1].d32 + ((v[1].d32 << 6) ^ (v[1].d32 >> 9))),
					prefetch[pi].rot1) + prefetch[pi].k1);
			}
			else
			{
				v[0].d32 -=
					(v[3].d32 +
					ror32 ((v[1].d32 + ((v[1].d32 << 6) ^ (v[1].d32 >> 9))) +
					prefetch[pi].k1, prefetch[pi].rot1));
			}


			v[0].d8[0] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[0]];
			v[0].d8[1] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[1]];
			v[0].d8[2] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[2]];
			v[0].d8[3] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[3]];

			v[1].d8[0] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[0]];
			v[1].d8[1] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[1]];
			v[1].d8[2] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[2]];
			v[1].d8[3] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[3]];

			v[2].d8[0] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[0]];
			v[2].d8[1] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[1]];
			v[2].d8[2] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[2]];
			v[2].d8[3] =
				BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[3]];

			v[3].d8[0] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[0]];
			v[3].d8[1] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[1]];
			v[3].d8[2] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[2]];
			v[3].d8[3] =
				BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[3]];

			s = v[0].d32 ^ v[1].d32 ^ v[2].d32 ^ v[3].d32;

			m = (s >> 24) & 3;
			t = v[3].d32;
			v[3].d32 = v[m].d32;
			v[m].d32 = t;

			m = (s >> 16) & 3;
			t = v[2].d32;
			v[2].d32 = v[m].d32;
			v[m].d32 = t;

			m = (s >> 8) & 3;
			t = v[1].d32;
			v[1].d32 = v[m].d32;
			v[m].d32 = t;

			m = s & 3;
			t = v[0].d32;
			v[0].d32 = v[m].d32;
			v[m].d32 = t;



		}
	}
}
